if not RazerNagaConfig then RazerNagaConfig = {} end

local debug = false
local T = LibStub( "WAR-AceLocale-3.0" ) : GetLocale( "RazerNaga", debug )

local windowId 					= "RazerNagaConfig"

local configWindowScroll		= windowId .. "Scroll"
local configWindowScrollChild	= configWindowScroll .. "Child"
local MAX_VISIBLE_ROWS 			= 15

-- Our list of registered windows
local configuredWindows			= {}

local displayOrder 				= {}		
RazerNagaConfig.displayData 			= {}

local currentDisplayedConfig	= 1

local dbOptions

local closeCallback

function RazerNagaConfig.GetWindowName() return windowId end
function RazerNagaConfig.IsShowing() return WindowGetShowing( windowId ) end
function RazerNagaConfig.GetActiveConfigWindowIndex() return currentDisplayedConfig end
function RazerNagaConfig.GetDBOptions() return dbOptions.handler end

function RazerNagaConfig.OnLoad( options, closeCB )
	-- Store our db options
	dbOptions = options
	
	closeCallback = closeCB

	-- Set our configuration title
	LabelSetText( windowId .. "TitleBarText", T["RazerNaga"] )
	
	-- Initialize our configuration windows
	for index, window in ipairs( configuredWindows )
    do
    	-- Create and store our window
    	window:Initialize()
    	
    	-- Hide the config dialog
    	RazerNagaConfig.HideConfig( index )
    end
    
    -- Display our initial window
    RazerNagaConfig.UpdateDisplayedConfig( currentDisplayedConfig )
    
    -- Revert all of our windows to set their settings
	RazerNagaConfig.RevertDialog()
end

function RazerNagaConfig.CreateDisplayData()
	-- Clear any existing data
	RazerNagaConfig.displayData = {}
	
	-- Create our list of configuration pages
	for slotNum, window in pairs( configuredWindows ) 
	do
		local listDisplayItem = {}
		
		listDisplayItem.slotNum 	= slotNum
		listDisplayItem.Name		= window.Name
    
    	RazerNagaConfig.displayData[slotNum] = listDisplayItem
	end
end

function RazerNagaConfig.UpdateConfigList()
	-- Clear our current display data
	displayOrder = {}
    
    -- Create the list we will use to display
    for index,_ in ipairs( RazerNagaConfig.displayData )
    do
    	-- Add this to the end of our display
    	table.insert( displayOrder, index )
    end
    
    -- Display the data
    ListBoxSetDisplayOrder( windowId .. "List", displayOrder )
end

function RazerNagaConfig.PopulateDisplay()
	local slotNum, config
	 
	if( RazerNagaConfigList.PopulatorIndices == nil ) then return end
	 
	for row, data in ipairs( RazerNagaConfigList.PopulatorIndices ) 
	do
		RazerNagaConfig.UpdateListRow( RazerNagaConfig.displayData[data], windowId .. "ListRow".. row )
	end 
end

function RazerNagaConfig.UpdateListRow( config, rowName )
	if( config.slotNum == currentDisplayedConfig ) then
		LabelSetTextColor( rowName .. "Name", DefaultColor.GREEN.r, DefaultColor.GREEN.g, DefaultColor.GREEN.b )
	else
		LabelSetTextColor( rowName .. "Name", 255 ,255 ,255 )
	end
end

function RazerNagaConfig.OnLButtonUpConfigList()
	local rowNumber, slotNumber, config = RazerNagaConfig.GetSlotRowNumForActiveListRow()
	
	if( slotNumber ~= currentDisplayedConfig ) then
		RazerNagaConfig.UpdateDisplayedConfig( slotNumber )	
	end
end

function RazerNagaConfig.UpdateDisplayedConfig( newConfig )
	local prevWindow = configWindowScrollChild
	
	-- Hide the old config
	RazerNagaConfig.HideConfig( currentDisplayedConfig )
	
	-- Clear the old config parent
	RazerNagaConfig.SetConfigParent( currentDisplayedConfig, "Root" )
	
	-- Set the parent appropriately
	RazerNagaConfig.SetConfigParent( newConfig, windowId )
	
	-- Anchor the configs children
	RazerNagaConfig.ReanchorConfigDisplay( newConfig, windowId )
	
	--
	-- Due to the requested removal of the scroll bar, only 1 dialog will be handled
	-- for display. 
	--
	local frame = configuredWindows[newConfig].display[1]
	
	if( frame ~= nil ) then
		frame:ClearAnchors()
		frame:AddAnchor( windowId, "top", "top", 20, 30 )
	end

	-- Display the new config
	RazerNagaConfig.ShowConfig( newConfig )
	
	-- Update our current displayed config
	currentDisplayedConfig = newConfig
end

function RazerNagaConfig.GetSlotRowNumForActiveListRow()
	local rowNumber, slowNumber, config
	
	-- Get the row within the window
	rowNumber = WindowGetId( SystemData.ActiveWindow.name ) 

	-- Get the data index from the list box
    local dataIndex = ListBoxGetDataIndex( windowId .. "List" , rowNumber )
    
    -- Get the slot from the data
    if( dataIndex ~= nil ) then
    	
    	slotNumber = RazerNagaConfig.displayData[dataIndex].slotNum
    
	    -- Get the data
	    if( slotNumber ~= nil ) then
	    	config = configuredWindows[slotNumber]
	    end
	end
    
	return slotNumber, rowNumber, config, window
end

function RazerNagaConfig.SaveDialog()
	-- Revert all of our windows to set their settings
	for index, window in ipairs( configuredWindows )
    do
    	window:Apply()
    end
end

function RazerNagaConfig.RevertDialog()
	-- Revert all of our windows to set their settings
	for index, window in ipairs( configuredWindows )
    do
    	window:Revert()
    end
end

function RazerNagaConfig.OnClose()
	WindowSetShowing( windowId, false )
end

function RazerNagaConfig.OnShown()

end

function RazerNagaConfig.OnHidden()
	if( closeCallback ~= nil and type( closeCallback ) == "function" ) then
		pcall( closeCallback )
	end
end

function RazerNagaConfig.OnApply()
	-- Save the dialog
	RazerNagaConfig.SaveDialog()
	
	-- Let the addon know the settings have changed
	RazerNaga.OnSettingsChanged()
end

function RazerNagaConfig.OnRevert()
	RazerNagaConfig.RevertDialog()
end

function RazerNagaConfig.RegisterWindow( window )
	table.insert( configuredWindows, window )
	return #configuredWindows
end

function RazerNagaConfig.OnResizeBegin()
    WindowUtils.BeginResize( windowId, "topleft", 800, 500, RazerNagaConfig.OnResizeEnd )
end

function RazerNagaConfig.OnResizeEnd()
	RazerNagaConfig.UpdateDisplayedConfig( currentDisplayedConfig )
end

function RazerNagaConfig.ShowConfig( index )
	if( configuredWindows[index] ~= nil and configuredWindows[index].display ~= nil ) then
		for k,v in ipairs( configuredWindows[index].display )
		do
			v:Show()
		end
	end
end

function RazerNagaConfig.HideConfig( index )
	if( configuredWindows[index] ~= nil and configuredWindows[index].display ~= nil ) then
		for k,v in ipairs( configuredWindows[index].display )
		do
			v:Hide()
		end
	end
end

function RazerNagaConfig.SetConfigParent( index, parent )
	if( configuredWindows[index] ~= nil and configuredWindows[index].display ~= nil ) then
		for k,v in ipairs( configuredWindows[index].display )
		do
			v:Parent( parent )
		end
	end
end

function RazerNagaConfig.ReanchorConfigDisplay( index, initialParent )
	local prevWindow	= initialParent
	local leftAnchor 	= "topleft"
	local rightAnchor 	= "topright"

	if( configuredWindows[index] ~= nil and configuredWindows[index].display ~= nil ) then
		for k,v in ipairs( configuredWindows[index].display )
		do
			v:ClearAnchors()
			v:AddAnchor( prevWindow, leftAnchor, "topleft", 0, 0 )
			v:AddAnchor( prevWindow, rightAnchor, "topright", 0, 0 )
			
			prevWindow = v.name
			leftAnchor = "bottomleft"
			rightAnchor = "bottomright"
		end
	end
end