--[[
	This application is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    The applications is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with the applications.  If not, see <http://www.gnu.org/licenses/>.
--]]

if RazerNaga == nil then RazerNaga = {} end

local pairs 		= pairs
local ipairs		= ipairs
local tonumber 		= tonumber
local tostring 		= tostring
local towstring 	= towstring
local tinsert		= table.insert
local tsort			= table.sort
local tconcat		= table.concat
local string_format = string.format

local alpha = false
local debug = false

local T 				= LibStub( "WAR-AceLocale-3.0" ) : GetLocale( "RazerNaga", debug )

local db				= {}

local Version 			= 010200   -- MM.mm.rr
local DisplayVersion 	= "v1.2.0"

local layoutEditorPrefix 	= "LayoutEditorWindowEdit"
local layoutEditorSuffix	= "Controls"

local actionBarLookup		= {}		-- Number to Name
local actionBarRevLookup	= {}		-- Name to Number

local rowsLookup			= {}		-- Index to Row Count for the right click menu
local preMouseOverRows		= 1			-- This stores the number of rows just prior to creating the row selection menu, to allow easy reverts

local iconId						= "RazerNagaMiniMap"

local baseContextMenu_RowsId		= "RazerNaga_ContextMenu_Rows"
local baseContextMenu_ModifierId	= "RazerNaga_ContextMenu_Modifier"

local firstLoad 				= true

local contextMenuFlag			= false
local contextMenuActionBarId	= nil

local ICON_CHECKED 		= L"<icon57>"
local ICON_UNCHECKED 	= L"<icon58>"
local BUTTON_COUNT		= 12

local INVERT_BUTTON_ORIENTATION_FALG	= 4
local FIRST_LOAD_BUTTON_COUNT_FLAG		= 12
local FIRST_LOAD_ROW_COUNT				= 4

local function rev_iter(t, key)
  if (key <= 1) then
    return nil
  else
    local nkey = key - 1
    return nkey, t[nkey]
  end
end

local function ripairs(t)
  return rev_iter, t, (1 + table.getn(t))
end

function RazerNaga.GetVersion()
	return DisplayVersion
end

function RazerNaga.OnInitialize()
	RegisterEventHandler( SystemData.Events.RELOAD_INTERFACE, "RazerNaga.OnLoad" )
	RegisterEventHandler( SystemData.Events.LOADING_END, "RazerNaga.OnLoad" )
end

function RazerNaga.OnLoad()
	if( firstLoad ) then
		-- Create our configuration objects
		RazerNaga.db	= LibStub( "WAR-AceDB-3.0" ) : New( "RazerNagaDatabase", RazerNaga.DefaultSettings, "Default" )
		db = RazerNaga.db.profile
		
		RazerNaga.dbOptions		= LibStub( "WAR-AceDBOptions-3.0" ) : GetOptionsTable( RazerNaga.db )
		RazerNaga.dbOptions.args = {} -- We do not use these in WAR
		
		-- Register a callback to know if the profile changes
		RazerNaga.db.RegisterCallback( RazerNaga, "OnProfileChanged", RazerNaga.OnProfileChanged )
		RazerNaga.db.RegisterCallback( RazerNaga, "OnProfileCopied", RazerNaga.OnProfileChanged )
		RazerNaga.db.RegisterCallback( RazerNaga, "OnProfileReset", RazerNaga.OnProfileChanged )

		--
		-- Build a lookup table of the action bar names
		--
		local i	= 1
		local bar = ActionBars:GetBar( i )
		while bar ~= nil
		do
			actionBarLookup[i] = bar:GetName()
			actionBarRevLookup[bar:GetName()] = i
			i = i + 1
			bar = ActionBars:GetBar( i )
		end
		
		-- Initialize our minimap icon
    	LayoutEditor.RegisterWindow( iconId, T["RazerNaga Toggle"], T["RazerNaga Toggle"], false, false, true, nil )
    	
    	-- Register for a layout editor finished event
    	LayoutEditor.RegisterEditCallback( RazerNaga.OnLayoutEditorEvent )
    	
    	-- Perform the auto keybind prompt check
		RazerNaga.CheckAutokeybindPrompt()
		
		-- Perform the first load check
		RazerNaga.CheckFirstLoad()
		
		-- Load our configuration window
		RazerNagaConfig.OnLoad( RazerNaga.dbOptions, RazerNaga.OnConfigClose )
		
		local oldEscapeKeyProcessed = WindowUtils.EscapeKeyProcessed
	    WindowUtils.EscapeKeyProcessed = function(...)
	        if( WindowGetShowing( RazerNagaConfig.GetWindowName() ) ) then
	        	RazerNagaConfig.OnClose()
	        	return
	        end
	        oldEscapeKeyProcessed(...)
    	end
    	
    	-- Handle our initial settings
		RazerNaga.OnSettingsChanged()
		
		-- Register our slash commands with LibSlash
		RazerNaga.Debug( "Registering slash commands" )
		
		-- Attempt to register one of our handlers
		if( LibSlash ~= nil and type( LibSlash.RegisterSlashCmd ) == "function" ) then
			LibSlash.RegisterSlashCmd( "razer", RazerNaga.Slash )
			LibSlash.RegisterSlashCmd( "razernaga", RazerNaga.Slash )
		end
		
		d( "RazerNaga loaded." )
	
		firstLoad = false
	end
end

function RazerNaga.OnSettingsChanged()
	-- Update our action bars
	local bar
	local count = #actionBarLookup
	for i = 1, count
	do
		bar = ActionBars:GetBar( i )
		if( bar ~= nil ) then
			RazerNaga.UpdateActionBar( bar, db.general.mode, db.bars[i], true )
		end
	end
end

function RazerNaga.UpdateActionBar( bar, mode, barSettings, updateKeybindings )
	-- Reanchor the bar
	RazerNaga.ReanchorActionBarButtons( bar, mode, barSettings )
	
	-- Update the layout editor anchor
	RazerNaga.ReanchorLayoutEditorFrame( bar )

	-- Update the keybindings if necessary
	if( updateKeybindings ) then
		RazerNaga.ReassignKeybindings( bar, mode, barSettings, false )
	end
end

function RazerNaga.OnLayoutEditorEvent( code )
	if( code == LayoutEditor.EDITING_BEGIN ) then
		-- Reanchor the layout editor frames
		local bar
		local count = #actionBarLookup
		for i = 1, count
		do
			bar = ActionBars:GetBar( i )
			if( bar ~= nil ) then
				RazerNaga.ReanchorLayoutEditorFrame( bar )
			end
		end
    end
end

function RazerNaga.OnProfileChanged( eventName )
	db 							= RazerNaga.db.profile
	RazerNaga.dbOptions			= LibStub( "WAR-AceDBOptions-3.0" ) : GetOptionsTable( RazerNaga.db )
	RazerNaga.dbOptions.args 	= {}
end

function RazerNaga.CheckFirstLoad()
	if( db.general.firstload == true ) then
		-- If the user's first bar is currently configured for 12 buttons, reshape the action bar to 3x4
		local buttonCount = ActionBarClusterManager.m_Settings:GetActionBarSetting( 1, "buttonCount" )
		
		if( buttonCount == FIRST_LOAD_BUTTON_COUNT_FLAG ) then
			RazerNaga.UpdateRows( 1, FIRST_LOAD_ROW_COUNT )
		end
		
		db.general.firstload = false
	end
end

function RazerNaga.CheckAutokeybindPrompt()
	-- Prompt the user to enable automatic keybinding if the prompt flag is true
	if( db.general.autokeybindingprompt == true ) then
		DialogManager.MakeTwoButtonDialog( T["RazerNaga - Do you wish to enable auto keybinding?"], 
			GetString( StringTables.Default.LABEL_YES ), RazerNaga.OnAutoKeybindingPrompt_Enable,
			GetString( StringTables.Default.LABEL_NO ), RazerNaga.OnAutoKeybindingPrompt_Disable )
		db.general.autokeybindingprompt = false
	end
end

function RazerNaga.OnAutoKeybindingPrompt_Enable()
	RazerNaga.UpdateAutomaticKeybinding( true )
	RazerNagaConfig.RevertDialog()
end

function RazerNaga.OnAutoKeybindingPrompt_Disable()
	RazerNaga.UpdateAutomaticKeybinding( false )
	RazerNagaConfig.RevertDialog()
end

function RazerNaga.ReanchorActionBarButtons( bar, mode, barSettings )
	if( bar == nil or mode == nil or barSettings == nil ) then
		d( "RazerNaga.ReanchorActionBarButtons received invalid parameter" ) 
		return 
	end
	
	--
	-- The goal of this function is to emulate ActionBar:AnchorButtons as close as possible,
	-- only modifying how the buttons are anchored.
	--
	-- It is based upon the 1.3 version of the function
	--
	if( bar.m_Buttons == nil ) then return end
	
	local buttonXPadding        = bar.m_ButtonXPadding
    local buttonXSpacing        = bar.m_ButtonXSpacing
    local buttonYPadding        = bar.m_ButtonYPadding
    local buttonYSpacing        = bar.m_ButtonYSpacing
    
    local buttons           	= bar.m_Buttons
    local buttonCount       	= #buttons
    local columnCount       	= bar.m_ColumnCount
    local windowName        	= bar:GetName() .. "Action"
    local rowCount 				= buttonCount / columnCount
    
    local buttonXSpacingCumulative 	= 0
    local buttonYSpacingCumulative 	= 0
    local newLineCount				= 0
    
    local anchor = {}
    local leftAnchor =
    {
        Point           = "topleft",
        RelativePoint   = "topleft",
        RelativeTo      = bar:GetName(),
        XOffset         = 0,
        YOffset         = 0,
    }
	local rightAnchor =
    {
        Point           = "bottomright",
        RelativePoint   = "bottomright",
        RelativeTo      = bar:GetName(),
        XOffset         = 0,
        YOffset         = 0,
    }
    
	if( columnCount == INVERT_BUTTON_ORIENTATION_FALG ) then 
		anchor =
		{
		    Point           = "bottomleft",
		    RelativePoint   = "bottomleft",
		    RelativeTo      = bar:GetName(),
		    XOffset         = buttonXPadding,
		    YOffset         = -buttonYPadding,
		}
		newLineCount = rowCount
	else
		anchor =
		{
		    Point           = "topleft",
		    RelativePoint   = "topleft",
		    RelativeTo      = bar:GetName(),
		    XOffset         = buttonXPadding,
		    YOffset         = buttonYPadding,
		}
		newLineCount = columnCount
	end
	
	for buttonIndex = 1, buttonCount
    do
        buttons[buttonIndex]:SetAnchor (anchor)
        
        local nextIndex = buttonIndex + 1
        local remainder = math.fmod( nextIndex, newLineCount )

	    if( ( remainder == 1 ) or ( newLineCount == 1 ) ) then
        	if( columnCount == INVERT_BUTTON_ORIENTATION_FALG ) then
	        	-- Start the new row to the right of the previous row start
	        	anchor.Point             = "topright"
	            anchor.RelativePoint     = "topleft"
	            anchor.RelativeTo        = windowName..(nextIndex - newLineCount)
	            anchor.XOffset           = buttonXSpacing
	            anchor.YOffset           = 0
	            buttonXSpacingCumulative = buttonXSpacingCumulative + buttonXSpacing
	        else
	        	-- Start the new row to the bottom of the previous row start
		        anchor.Point             = "bottomleft"
	            anchor.RelativePoint     = "topleft"
	            anchor.RelativeTo        = windowName..(nextIndex - newLineCount)
	            anchor.XOffset           = 0
	            anchor.YOffset           = buttonYSpacing
	            buttonYSpacingCumulative = buttonYSpacingCumulative + buttonYSpacing
	        end
        else
        	if( columnCount == INVERT_BUTTON_ORIENTATION_FALG ) then
	        	-- Anchor this button above its predecessor
	        	anchor.Point             = "top"
	            anchor.RelativePoint     = "bottom"
	            anchor.RelativeTo        = windowName..buttonIndex
	            anchor.XOffset           = 0
	            anchor.YOffset           = buttonYSpacing
	            buttonYSpacingCumulative = buttonYSpacingCumulative + buttonYSpacing
	        else
	        	-- Anchor this button to the right side of its predecessor
	            anchor.Point             = "right"
	            anchor.RelativePoint     = "left"
	            anchor.RelativeTo        = windowName..buttonIndex
	            anchor.XOffset           = buttonXSpacing
	            anchor.YOffset           = 0
	            buttonXSpacingCumulative = buttonXSpacingCumulative + buttonXSpacing
	        end
        end
    end    
    
    local buttonWidth, buttonHeight = buttons[1]:GetDimensions()
    local pageSelectorWidth = 0
    if( bar.m_PageSelectorWindow and bar.m_PageSelectorMode ~= ActionBarConstants.HIDE_PAGE_SELECTOR )
    then
        pageSelectorWidth, _ = bar.m_PageSelectorWindow:GetDimensions()
    else
        pageSelectorWidth = buttonXPadding
    end
    
    local finalWidth  = (columnCount * buttonWidth)  + buttonXSpacingCumulative + buttonXPadding + pageSelectorWidth
    local finalHeight = (rowCount    * buttonHeight) + buttonYSpacingCumulative + (buttonYPadding * 2)
    
    bar:SetDimensions( finalWidth, finalHeight, Frame.FORCE_OVERRIDE )
    
    if (bar.m_PageSelectorWindow ~= nil)
    then
        if (bar.m_PageSelectorMode == ActionBarConstants.SHOW_PAGE_SELECTOR_RIGHT)
        then           
            bar.m_PageSelectorWindow:SetAnchor ({Point = "topright", RelativePoint = "topleft", RelativeTo = windowName..buttonCount, XOffset = -2, YOffset = -2})
        elseif (bar.m_PageSelectorMode == ActionBarConstants.SHOW_PAGE_SELECTOR_LEFT)
        then
            bar.m_PageSelectorWindow:SetAnchor ({Point = "left", RelativePoint = "left", RelativeTo = bar:GetName (), XOffset = buttonXPadding, YOffset = buttonYPadding - 2})
            buttons[1]:SetAnchor ({Point = "right", RelativePoint = "left", RelativeTo = bar.m_PageSelectorWindow:GetName (), XOffset = 0, YOffset = 0})
        end
    end    
    
    -- Force Process the anchors on the bar so everything looks right
    bar:ForceProcessAnchors ()
    
    -- Reanchor the background
    bar.m_Background:SetAnchor( leftAnchor, rightAnchor )
end

function RazerNaga.ReassignKeybindings( bar, mode, barSettings, bypass )
	if( bar == nil or mode == nil or barSettings == nil ) then
		d( "RazerNaga.ReassignKeybindings received invalid parameter" ) 
		return 
	end
	
	if( bypass == false and ( not db.general.automatickeybinding or not barSettings.automatickeybinding ) ) then return end
	
	local bindings
	
	if( mode == RazerNaga.MODE_BASIC ) then
		bindings = db.bindings.basic
	elseif( mode == RazerNaga.MODE_ADVANCED ) then
		bindings = db.bindings.advanced
	else
		d( "RazerNaga.ReassignKeybindings unhandled mode." )
		return
	end
	
	for idx, button in ipairs( bar.m_Buttons )
	do
		local data = bindings[idx]
	
		-- Make sure we received data and that its a table	
		if( data ~= nil and type(data) == "table" ) then
			local currentBindings = {}
			local oldBinding
			
			-- Get the current keybindings
			KeyUtils.GetBindingsForAction( button.m_ActionName, currentBindings )
			
			-- Clear the current keybindings
			RemoveAllBindings( button.m_ActionName )
			
			-- Prefer to keep the later most binding
			for _, binding in ripairs( currentBindings )
			do
				if( binding ~= nil ) then
					oldBinding = binding
					break
				end
			end
			
			-- If the modifier is set to unassigned, leave the button w/o any binding
			if( barSettings.modifier ~= RazerNaga.MODIFIER_UNASSIGNED ) then
				-- If there was an old binding add it to the button
				if( oldBinding ~= nil ) then
					AddBinding( button.m_ActionName, oldBinding.deviceId, oldBinding.buttons )
				end
			
				-- Add the appropriate bindings from our configuration
				for _, binding in ipairs( data.bindings ) 
				do
					local buttons = DataUtils.CopyTable( binding.buttons )
					
					local modButton = RazerNaga.MODIFIER[barSettings.modifier].button
					if( modButton ~= nil and modButton ~= 0 ) then
						tinsert( buttons, 1, modButton )
					end
					AddBinding( button.m_ActionName, binding.deviceId, buttons )
	        	end
	        end
        	
        	-- Update the keybinding text for the button
    		button:UpdateKeyBindingText()
		end
	end
	
	-- Update the keymapping window accordingly
	KeyMappingWindow.UpdateData()
end

function RazerNaga.ToggleConfig()
	if( DoesWindowExist( RazerNagaConfig.GetWindowName() ) ) then
		-- Make sure we can get the action bar
		if( not WindowGetShowing( RazerNagaConfig.GetWindowName() ) ) then
			-- Begin the layout editor
			LayoutEditor.Begin()
			
			-- Start editing mode
			LayoutEditor.StartEditing()
		
			-- Hide the HUD Controls
			if( DoesWindowExist( LayoutEditor.WINDOW_NAME .. "HUDControls" ) ) then
				WindowSetShowing( LayoutEditor.WINDOW_NAME .. "HUDControls", false )
			end
			
			-- Display our configuration
			WindowSetShowing( RazerNagaConfig.GetWindowName(), true )
		end
	end
end

function RazerNaga.LayoutEditor_RestoreDefaults(...)
	-- Call the old hook
	RazerNaga.__LayoutEditor_RestoreDefaults(...)
	
	local nonRestore = {}
	local defaultAutomaticKeybinding = db.general.automatickeybinding
	
	-- Make a backup of non restoring configuration members
	for k, v in pairs( db.bars )
	do
		if( nonRestore[k] == nil ) then
			nonRestore[k] = {}
		end
		
		nonRestore[k].automatickeybinding = v.automatickeybinding
		nonRestore[k].modifier = v.modifier
	end

	-- Set the temp profile active
	RazerNaga.db:SetProfile( "temp" )
	
	-- Reset the temp profile
	RazerNaga.db:ResetProfile()
	
	-- Restore the backup of our non restoring configuration members
	for k, v in pairs( nonRestore )
	do
		db.bars[k].automatickeybinding 	= v.automatickeybinding
		db.bars[k].modifier 			= v.modifier
	end
	db.general.autokeybindingprompt 	= false
	db.general.automatickeybinding 		= defaultAutomaticKeybinding
	db.general.firstload				= false
	
	local bar
	local count = #actionBarLookup
	for i = 1, count
	do
		bar = ActionBars:GetBar( i )
		if( bar ~= nil ) then
			RazerNaga.UpdateActionBar( bar, db.general.mode, db.bars[i], false )
		end
	end
end

function RazerNaga.LayoutEditor_SaveExit(...)
	-- Call the old hook
	RazerNaga.__LayoutEditor_SaveExit(...)
	
	-- Set the default profile active
	RazerNaga.db:SetProfile( "Default" )
	
	-- If the temp profile exists, copy the profile into ours and delete it
	local profiles = RazerNaga.dbOptions.handler:ListProfiles( { arg="nocurrent" } )
	for k, v in pairs( profiles )
	do
		if( k == "temp" ) then
			-- Copy temp profile into the default profile
			RazerNaga.db:CopyProfile( "temp" )
			
			-- Delete the temp profile as we no longer need it
			RazerNaga.db:DeleteProfile( "temp" )
			break
		end
	end
	
	-- Flag that our settings have updated
	RazerNaga.OnSettingsChanged()
end

function RazerNaga.LayoutEditor_ResetExit(...)
	-- Call the old hook
	RazerNaga.__LayoutEditor_ResetExit(...)
	
	-- Set the default profile active
	RazerNaga.db:SetProfile( "Default" )
	
	-- Delete the temp profile as we no longer need it
	RazerNaga.db:DeleteProfile( "temp" )
	
	-- Update our settings
	RazerNaga.OnSettingsChanged()
end

function RazerNaga.SettingsWindowTabbed_OnResetButton(...)
	-- Call the old hook
	RazerNaga.__SettingsWindowTabbed_OnResetButton(...)	
	RazerNaga.OnSettingsChanged()
end

function RazerNaga.SettingsWindowTabInterface_ApplyCurrent(...)
	-- Call the old hook
	RazerNaga.__SettingsWindowTabInterface_ApplyCurrent(...)
	RazerNaga.OnSettingsChanged()
end

function RazerNaga.SettingsWindowTabInterface_SelectRows(...)
	-- Call the old hook
	RazerNaga.__SettingsWindowTabInterface_SelectRows(...)
	
	local barNum = WindowGetId( WindowGetParent( SystemData.ActiveWindow.name ) )
	local bar = ActionBars:GetBar( barNum )
	if( bar ~= nil ) then
		RazerNaga.UpdateActionBar( bar, db.general.mode, db.bars[barNum], true )
	end
end

function RazerNaga.ReanchorLayoutEditorFrame( actionBar )
	--
	-- Because the user is able to change the orientation of the action bar while within the editor
	-- the Layout Edit window and control window that gets overlayed the action bar needs to be reanchored from it single
	-- anchor point top left, to a topleft/bottomright anchor
	-- ex: "LayoutEditorWindowEditEA_ActionBar_1"
	--
	local windowId 	= LayoutEditor.WINDOW_NAME .. "Edit" .. actionBar:GetName()
	local frame 	= FrameManager:Get( windowId )
	
	-- Make sure we were able to get the action frame
	if( frame ~= nil ) then
		local topLeft = 
		{
			RelativeTo 		= actionBar:GetName(),
			Point			= "topleft",
			RelativePoint	= "topleft",
			XOffset			= 0,
			YOffset			= 0,
		}
		
		local bottomRight = 
		{
			RelativeTo 		= actionBar:GetName(),
			Point			= "bottomright",
			RelativePoint	= "bottomright",
			XOffset			= 0,
			YOffset			= 0,
		}
				
		-- Reanchor the frame 
		frame:SetAnchor( topLeft, bottomRight )
		
		-- Attempt to retrieve the controls window for this layout editor frame
		local controlFrame 	= FrameManager:Get( windowId .. "Controls" )
		
		-- This frame will only exist if the user has it as the active window in the layout editor
		if( controlFrame ~= nil ) then
			-- Have the frame reattach itself to the proper frame
			controlFrame:Attach()
		end
	end
end

function RazerNaga.OnConfigClose()
	if( LayoutEditor.isActive ) then
		-- Let the layout editor know to exit
		LayoutEditor.SaveExit()
	end
	
	-- Show the HUD controls window
	if( DoesWindowExist( LayoutEditor.WINDOW_NAME .. "HUDControls" ) ) then
		WindowSetShowing( LayoutEditor.WINDOW_NAME .. "HUDControls", true )
	end
end

function RazerNaga.Slash( input )
	RazerNaga.ToggleConfig()
end

function RazerNaga.Print( wstr )
	EA_ChatWindow.Print( towstring( wstr ), ChatSettings.Channels[SystemData.ChatLogFilters.SAY].id )
end

function RazerNaga.Debug(str)
	if( db ~= nil and db.general.debug ) then
	 	DEBUG(towstring(str))
	end
end

function RazerNaga.EA_Window_ContextMenu_CreateContextMenu( ... )
	-- Execute the hook first
	RazerNaga.__EA_Window_ContextMenu_CreateContextMenu( ... )
	
	-- Reset the context menu flag
	contextMenuFlag 		= false
		
	-- Only process this if the layout editor is active, and configuration window is displayed
	if( LayoutEditor.isActive and DoesWindowExist( RazerNagaConfig.GetWindowName() ) and WindowGetShowing( RazerNagaConfig.GetWindowName() ) ) then
		local window = SystemData.ActiveWindow.name:match( layoutEditorPrefix .. "(.+)" .. layoutEditorSuffix )
		
		-- Set our context menu flag
		contextMenuFlag = ( actionBarRevLookup[window] ~= nil )
		
		-- Only update our bar Id if its nil, otherwise we may pull the ID out from under a cascading menu
		if( contextMenuActionBarId == nil ) then
			contextMenuActionBarId = actionBarRevLookup[window]
		end
	end
end

function RazerNaga.EA_Window_ContextMenu_Finalize( ... )
	-- If we are context menu flagged, add our menu options
	if( LayoutEditor.isActive and contextMenuFlag == true ) then
		-- Add the RazerNaga menu items
		EA_Window_ContextMenu.AddMenuDivider( EA_Window_ContextMenu.CONTEXT_MENU_1 )
		EA_Window_ContextMenu.AddCascadingMenuItem( T["RazerNaga"], RazerNaga.OnContextMenu_Base, false, EA_Window_ContextMenu.CONTEXT_MENU_1 )
		
		-- Unflag our context
		contextMenuFlag = false
	end
	
	-- Execute the hook
	RazerNaga.__EA_Window_ContextMenu_Finalize( ... )
end

function RazerNaga.EA_Window_Context_Menu_HideAll()
	-- Call our base class first
	RazerNaga.__EA_Window_ContextMenu_HideAll()
	
	-- Once the menu closes, we can nil our bar id
	contextMenuActionBarId = nil
end

function RazerNaga.OnContextMenu_Base()
	EA_Window_ContextMenu.CreateContextMenu( "", EA_Window_ContextMenu.CONTEXT_MENU_2 )
	
	local icon = ICON_UNCHECKED
	if( db.bars[contextMenuActionBarId].automatickeybinding ) then
		icon = ICON_CHECKED
	end
	EA_Window_ContextMenu.AddMenuItem( icon .. L"  " .. T["Enable Automatic Keybinding"], 
			function()
				RazerNaga.UpdateAutomaticKeybindingBar( contextMenuActionBarId )
			end, 
			false, true, EA_Window_ContextMenu.CONTEXT_MENU_2 )
	
	EA_Window_ContextMenu.AddMenuDivider( EA_Window_ContextMenu.CONTEXT_MENU_2 )
	EA_Window_ContextMenu.AddCascadingMenuItem( T["Modifier"], RazerNaga.OnContextMenu_Modifier, false, EA_Window_ContextMenu.CONTEXT_MENU_2 )
	EA_Window_ContextMenu.AddCascadingMenuItem( GetStringFromTable( "UserSettingsStrings", StringTables.UserSettings.LABEL_ACTION_BAR_ROWS ), RazerNaga.OnContextMenu_Rows, false, EA_Window_ContextMenu.CONTEXT_MENU_2 )
	
	EA_Window_ContextMenu.Finalize( EA_Window_ContextMenu.CONTEXT_MENU_2 )
end

function RazerNaga.OnContextMenu_Rows()
	EA_Window_ContextMenu.CreateContextMenu( "", EA_Window_ContextMenu.CONTEXT_MENU_3 )
	
	local buttonCount 	= ActionBarClusterManager.m_Settings:GetActionBarSetting( contextMenuActionBarId, "buttonCount" )
	local columns		= ActionBarClusterManager.m_Settings:GetActionBarSetting( contextMenuActionBarId, "columns" )
	
	-- Store our current row count (incase we need it on a mouse over end to reset the display)
	preMouseOverRows = buttonCount / columns
	
	-- Clear our rows
	rowsLookup = {}
	
	local numButtons = buttonCount
    if( numButtons < 1 ) then
        numButtons = 1
    elseif( numButtons > ActionBarConstants.BUTTONS ) then
        numButtons = ActionBarConstants.BUTTONS
    end
	
	local halfButtons = math.floor( buttonCount / 2 )
	local indexCount = 1
	for j = 1, halfButtons
    do
        if( math.mod( numButtons, j ) == 0 ) then
        	rowsLookup[indexCount] = j
            EA_Window_ContextMenu.AddUserDefinedMenuItem( createContextMenu_RowsWindow( indexCount, towstring( j ), j == preMouseOverRows ), EA_Window_ContextMenu.CONTEXT_MENU_3 )
            indexCount = indexCount + 1
        end
    end
    rowsLookup[indexCount] = buttonCount
    EA_Window_ContextMenu.AddUserDefinedMenuItem( createContextMenu_RowsWindow( indexCount, towstring( buttonCount ), buttonCount == preMouseOverRows ), EA_Window_ContextMenu.CONTEXT_MENU_3 )
    EA_Window_ContextMenu.Finalize( EA_Window_ContextMenu.CONTEXT_MENU_3 )
end

function RazerNaga.OnContextMenu_Modifier()
	EA_Window_ContextMenu.CreateContextMenu( "", EA_Window_ContextMenu.CONTEXT_MENU_3 )
	
	local data
	local bar = db.bars[contextMenuActionBarId]
	
	if( bar ~= nil ) then
		for idx, key in ipairs( RazerNaga.MODIFIER_MENU )
		do
			data = RazerNaga.MODIFIER[key]
			if( data ~= nil ) then
				EA_Window_ContextMenu.AddUserDefinedMenuItem( createContextMenu_ModifierWindow( idx, data.text, bar.modifier == key ), EA_Window_ContextMenu.CONTEXT_MENU_3 )
			end
		end
	end
    EA_Window_ContextMenu.Finalize( EA_Window_ContextMenu.CONTEXT_MENU_3 )
end

function RazerNaga.OnMouseOver_ContextMenu_Rows()
	local id = WindowGetId( SystemData.MouseOverWindow.name )
	if( id == nil ) then return end
	local rows = rowsLookup[id]
	if( rows ) then
		RazerNaga.UpdateRows( contextMenuActionBarId, rows )
	end
end

function RazerNaga.OnMouseOverEnd_ContextMenu_Rows()
	if( contextMenuActionBarId ~= nil ) then
		RazerNaga.UpdateRows( contextMenuActionBarId, preMouseOverRows )
	end
end

function RazerNaga.OnLButtonUp_ContextMenu_Rows()
	local id = WindowGetId( SystemData.MouseOverWindow.name )
	if( id == nil  ) then return end
	local rows = rowsLookup[id]
	if( rows ) then
		RazerNaga.UpdateRows( contextMenuActionBarId, rows )
	end
	EA_Window_ContextMenu.HideAll()
end

function RazerNaga.OnLButtonUp_ContextMenu_Modifier()
	local id = WindowGetId( SystemData.MouseOverWindow.name )
	local modifier = RazerNaga.MODIFIER[RazerNaga.MODIFIER_MENU[id]]
	
	if( modifier.update ~= nil ) then
		modifier.update( contextMenuActionBarId )
	end
	EA_Window_ContextMenu.HideAll()
end

function createContextMenu_RowsWindow( index, text, disabled )
	local windowId = baseContextMenu_RowsId .. index
	
	-- Check if we need to create the window, or reuse the old
	if( not DoesWindowExist( windowId ) ) then
		CreateWindowFromTemplate( windowId, baseContextMenu_RowsId, "Root" )
		WindowSetId( windowId, index )
	end
	
	-- Populate the items
	ButtonSetText( windowId, text )
	ButtonSetDisabledFlag( windowId, disabled )
	
	return windowId
end

function createContextMenu_ModifierWindow( index, text, disabled )
	local windowId = baseContextMenu_ModifierId .. index
	
	-- Check if we need to create the window, or reuse the old
	if( not DoesWindowExist( windowId ) ) then
		CreateWindowFromTemplate( windowId, baseContextMenu_ModifierId, "Root" )
		WindowSetId( windowId, index )
	end
	
	-- Populate the items
	ButtonSetText( windowId, text )
	ButtonSetDisabledFlag( windowId, disabled )
	
	return windowId
end

function RazerNaga.UpdateRows( bar, rows )
	local actionBar = ActionBars:GetBar( bar )
	if( actionBar ~= nil ) then
		local cols = ActionBarClusterManager.m_Settings:GetActionBarSetting( bar, "buttonCount" ) / rows
		
		-- We emulate most of the functionality within:
		-- 		SettingsWindowTabInterface.SetActionBarsCols
		ActionBarClusterManager.m_Settings:SetActionBarSetting( bar, "columns", cols )
    	ActionBars.m_Bars[bar].m_ColumnCount = cols
    	local rows = ActionBarClusterManager.m_Settings:GetActionBarSetting( bar, "buttonCount" ) / cols
		local oldMode = ActionBarClusterManager.m_Settings:GetActionBarSetting( bar, "selector" )
		local newMode = ActionBarConstants.SHOW_PAGE_SELECTOR_RIGHT
		
		if( cols <= rows ) then newMode = ActionBarConstants.HIDE_PAGE_SELECTOR	end
		if( oldMode ~= newMode ) then
		    ActionBarClusterManager.m_Settings:SetActionBarSetting( bar, "selector", newMode )
		    ActionBars.m_Bars[bar]:SetPageSelectorMode( newMode )
		end
    
        RazerNaga.UpdateActionBar( actionBar, db.general.mode, db.bars[bar], false )
    end
end

function RazerNaga.UpdateMode( mode )
	db.general.mode = mode
	RazerNaga.OnSettingsChanged()
end

function RazerNaga.UpdateAutomaticKeybinding( on )
	db.general.automatickeybinding = on
	RazerNaga.OnSettingsChanged()
end

function RazerNaga.UpdateAutomaticKeybindingBar( bar )
	db.bars[bar].automatickeybinding = not db.bars[bar].automatickeybinding
	local actionBar = ActionBars:GetBar( bar )
	if( actionBar ~= nil ) then
		RazerNaga.UpdateActionBar( actionBar, db.general.mode, RazerNaga.ORIENTATION[db.bars[bar].orientation], db.bars[bar], true )
	end
end

function RazerNaga.UpdateModifier( bar, modifier )
	local actionBar
	-- If there are any other bar with this modifier, set their modifier to NONE
	for k, v in pairs( db.bars )
	do
		if( k ~= bar and v.modifier == modifier ) then
			v.modifier = RazerNaga.MODIFIER_UNASSIGNED
			actionBar = ActionBars:GetBar( k )
			RazerNaga.ReassignKeybindings( actionBar, db.general.mode, v, true )
		end
	end
	db.bars[bar].modifier = modifier
	actionBar = ActionBars:GetBar( bar )
	RazerNaga.ReassignKeybindings( actionBar, db.general.mode, db.bars[bar], true )
end

--
-- Hook these 2 functions to determine when a user right clicks
--
RazerNaga.__EA_Window_ContextMenu_CreateContextMenu = EA_Window_ContextMenu.CreateContextMenu
EA_Window_ContextMenu.CreateContextMenu = RazerNaga.EA_Window_ContextMenu_CreateContextMenu

RazerNaga.__EA_Window_ContextMenu_Finalize = EA_Window_ContextMenu.Finalize
EA_Window_ContextMenu.Finalize = RazerNaga.EA_Window_ContextMenu_Finalize

RazerNaga.__EA_Window_ContextMenu_HideAll = EA_Window_ContextMenu.HideAll
EA_Window_ContextMenu.HideAll = RazerNaga.EA_Window_Context_Menu_HideAll

RazerNaga.__LayoutEditor_RestoreDefaults			= LayoutEditor.RestoreDefaults
LayoutEditor.RestoreDefaults 						= RazerNaga.LayoutEditor_RestoreDefaults

RazerNaga.__LayoutEditor_SaveExit					= LayoutEditor.SaveExit
LayoutEditor.SaveExit 								= RazerNaga.LayoutEditor_SaveExit

RazerNaga.__LayoutEditor_ResetExit					= LayoutEditor.ResetExit
LayoutEditor.ResetExit								= RazerNaga.LayoutEditor_ResetExit

RazerNaga.__SettingsWindowTabbed_OnResetButton		= SettingsWindowTabbed.OnResetButton
SettingsWindowTabbed.OnResetButton					= RazerNaga.SettingsWindowTabbed_OnResetButton

RazerNaga.__SettingsWindowTabInterface_ApplyCurrent	= SettingsWindowTabInterface.ApplyCurrent
SettingsWindowTabInterface.ApplyCurrent				= RazerNaga.SettingsWindowTabInterface_ApplyCurrent

RazerNaga.__SettingsWindowTabInterface_SelectRows	= SettingsWindowTabInterface.SelectRows
SettingsWindowTabInterface.SelectRows				= RazerNaga.SettingsWindowTabInterface_SelectRows 